"""
Utility functions for the multigrid solver.
"""
# Copyright 2018-2021 The EMSiG community.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import copy
import warnings
import importlib
from time import perf_counter
from datetime import datetime, timedelta
from concurrent.futures import ProcessPoolExecutor

import numpy as np

try:
    from scooby import Report as ScoobyReport
except ImportError:
    class ScoobyReport:
        """Dummy placeholder."""

try:
    import tqdm
    import tqdm.contrib.concurrent
except ImportError:
    tqdm = None

# Version: We take care of it here instead of in __init__, so we can use it
# within the package itself (logs).
try:
    # - Released versions just tags:       0.8.0
    # - GitHub commits add .dev#+hash:     0.8.1.dev4+g2785721
    # - Uncommitted changes add timestamp: 0.8.1.dev4+g2785721.d20191022
    from emg3d.version import version as __version__
except ImportError:
    # If it was not installed, then we don't know the version. We could throw a
    # warning here, but this case *should* be rare. emg3d should be installed
    # properly!
    __version__ = 'unknown-'+datetime.today().strftime('%Y%m%d')

__all__ = ['Report', 'EMArray', 'Timer']


# PRIVATE UTILS
_KNOWN_CLASSES = {}  # List of known classes for (de-)serialization


def _known_class(func):
    """Decorator to register class as known for I/O."""
    _KNOWN_CLASSES[func.__name__] = func
    return func


def _requires(*args, **kwargs):
    """Decorator to wrap functions with extra dependencies.

    This function is taken from `pysal` (in `lib/common.py`); see
    https://github.com/pysal/pysal (released under the 'BSD 3-Clause "New" or
    "Revised" License').


    Parameters
    ---------
    args : list
        Strings containing the modules to import.

    requires_verbose : bool
        If True (default) print a warning message on import failure.


    Returns
    -------
    out : func
        Original function if all modules are importable, otherwise returns a
        function that passes.

    """
    def simport(modname):
        """Safely import a module without raising an error."""
        try:
            return True, importlib.import_module(modname)
        except ImportError:
            return False, None

    def inner(function):
        available = [simport(arg)[0] for arg in args]
        if all(available):
            return function
        else:
            verbose = kwargs.pop('requires_verbose', True)
            wanted = copy.deepcopy(args)

            def passer(*args, **kwargs):
                if verbose:
                    missing = [arg for i, arg in enumerate(wanted)
                               if not available[i]]
                    # Print is always shown and simpler, warn for the CLI logs.
                    msg = ("This feature of emg3d requires the missing "
                           f"soft dependencies {missing}.")
                    print(f"* WARNING :: {msg}")
                    warnings.warn(msg, UserWarning)
            return passer

    return inner


def _process_map(fn, *iterables, max_workers, **kwargs):
    """Dispatch processes in parallel or not, using tqdm or not.

    :class:`emg3d.simulations.Simulation` uses the function
    ``tqdm.contrib.concurrent.process_map`` to run jobs asynchronously.
    However, ``tqdm`` is a soft dependency. In case it is not installed we use
    the class ``concurrent.futures.ProcessPoolExecutor`` directly, from the
    standard library, and imitate the behaviour of process_map (basically a
    ``ProcessPoolExecutor.map``, returned as a list, and wrapped in a context
    manager). If max_workers is smaller than two then we we avoid parallel
    execution.

    """
    # Parallel
    if max_workers > 1 and tqdm is None:
        with ProcessPoolExecutor(max_workers=max_workers) as ex:
            return list(ex.map(fn, *iterables))

    # Parallel with tqdm
    elif max_workers > 1:
        return tqdm.contrib.concurrent.process_map(
                fn, *iterables, max_workers=max_workers, **kwargs)

    # Sequential
    elif tqdm is None:
        return list(map(fn, *iterables))

    # Sequential with tqdm
    else:
        return list(tqdm.auto.tqdm(
            iterable=map(fn, *iterables), total=len(iterables[0]), **kwargs))


# PUBLIC UTILS
@_requires('scooby')
class Report(ScoobyReport):
    r"""Print date, time, and version information.

    Use ``scooby`` to print date, time, and package version information in any
    environment (Jupyter notebook, IPython console, Python console, QT
    console), either as html-table (notebook) or as plain text (anywhere).

    Always shown are the OS, number of CPU(s), ``numpy``, ``scipy``, ``emg3d``,
    ``numba``, ``sys.version``, and time/date.

    Additionally shown are, if they can be imported, ``IPython``,
    ``matplotlib``, and all soft dependencies of ``emg3d``. It also shows MKL
    information, if available.

    All modules provided in ``add_pckg`` are also shown.

    .. note::

        The package ``scooby`` has to be installed in order to use ``Report``:
        ``pip install scooby`` or ``conda install -c conda-forge scooby``.


    Parameters
    ----------
    add_pckg : {package, str}, default: None
        Package or list of packages to add to output information (must be
        imported beforehand or provided as string).

    ncol : int, default: 3
        Number of package-columns in html table (no effect in text-version).

    text_width : int, default: 80
        The text width for non-HTML display modes

    sort : bool, default: False
        Sort the packages when the report is shown

    """

    def __init__(self, add_pckg=None, ncol=3, text_width=80, sort=False):
        """Initiate a scooby.Report instance."""

        # Mandatory packages.
        core = ['numpy', 'scipy', 'numba', 'emg3d']

        # Optional packages.
        optional = ['empymod', 'xarray', 'discretize', 'h5py', 'matplotlib',
                    'tqdm', 'IPython']

        super().__init__(additional=add_pckg, core=core, optional=optional,
                         ncol=ncol, text_width=text_width, sort=sort)


class EMArray(np.ndarray):
    r"""An EM-ndarray adds the methods `amp` (amplitude) and `pha` (phase).

    Parameters
    ----------
    data : ndarray
        Data to which to add ``.amp`` and ``.pha`` attributes.


    Examples
    --------

    .. ipython::

       In [1]: import numpy as np
          ...: from empymod.utils import EMArray
          ...: emvalues = EMArray(np.array([1+1j, 1-4j, -1+2j]))

       # Amplitudes
       In [2]: emvalues.amp()
       Out[2]: EMArray([1.41421356, 4.12310563, 2.23606798])

       # Phase in radians
       In [3]: emvalues.pha()
       Out[3]: EMArray([ 0.78539816, -1.32581766, -4.24874137])

       # Phase in degrees
       In [4]: emvalues.pha(deg=True)
       Out[4]: EMArray([  45.        ,  -75.96375653, -243.43494882])

       # Phase in degrees, lead defined
       In [5]: emvalues.pha(deg=True, lag=False)
       Out[5]: EMArray([-45.        ,  75.96375653, 243.43494882])

    """

    def __new__(cls, data):
        r"""Create a new EMArray."""
        return np.asarray(data).view(cls)

    def amp(self):
        """Amplitude of the electromagnetic field."""
        return np.abs(self.view())

    def pha(self, deg=False, unwrap=True, lag=True):
        """Phase of the electromagnetic field.

        Parameters
        ----------
        deg : bool, default: False
            The returned phase is in degrees if True, else in radians.

        unwrap : bool, default: True
            The returned phase is unwrapped if True.

        lag : bool, default: True
            The returned phase is lag defined if True, else lead defined.

        """
        # Get phase, lead or lag defined.
        if lag:
            pha = np.angle(self.view())
        else:
            pha = np.angle(np.conj(self.view()))

        # Unwrap if `unwrap`.
        # np.unwrap removes the EMArray class;
        # for consistency, we wrap it in EMArray again.
        if unwrap and self.size > 1:
            pha = EMArray(np.unwrap(pha))

        # Convert to degrees if `deg`.
        if deg:
            pha *= 180/np.pi

        return pha


class Timer:
    """Class for timing (now; runtime)."""

    def __init__(self):
        """Initiate timer with a performance counter."""
        self._t0 = perf_counter()

    def __repr__(self):
        """Simple representation."""
        return f"Runtime : {self.runtime}"

    @property
    def t0(self):
        """Return time zero of this class instance."""
        return self._t0

    @property
    def now(self):
        """Return current time as hh:mm:ss string."""
        return datetime.now().strftime("%H:%M:%S")

    @property
    def runtime(self):
        """Return elapsed time as hh:mm:ss string."""
        return str(timedelta(seconds=np.round(self.elapsed)))

    @property
    def elapsed(self):
        """Return elapsed time in seconds."""
        return perf_counter() - self._t0



"""
Everything related to meshes appropriate for the multigrid solver.
"""
# Copyright 2018-2021 The EMSiG community.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

from copy import deepcopy

import numpy as np
from scipy.constants import mu_0

from emg3d import maps, utils

try:
    import discretize
except ImportError:
    discretize = None

__all__ = ['TensorMesh', 'BaseMesh', 'construct_mesh', 'origin_and_widths',
           'good_mg_cell_nr', 'skin_depth', 'wavelength', 'cell_width']


class BaseMesh:
    """Minimal TensorMesh for internal multigrid computation.

    The base mesh has everything that is needed within
    :func:`emg3d.solver.solve`, but nothing more.

    Parameters
    ----------
    h : [array_like, array_like, array_like]
        Cell widths in x, y, and z directions.

    origin : array_like
        Origin (x, y, z).


    Examples
    --------

    .. ipython::

       In [1]: import emg3d
          ...: import numpy as np

       In [2]: # Create a simple grid, 8 cells of length 100 m in each
          ...: # direction, centered around the origin.
          ...: hx = np.ones(8)*100
          ...: grid = emg3d.meshes.BaseMesh(
          ...:            [hx, hx, hx], origin=(-400, -400, -400))
          ...: grid  # QC grid

    """

    def __init__(self, h, origin, **kwargs):
        """Initialize the mesh."""

        # Store origin.
        self.origin = np.array(origin)

        # Width of cells, cast to arrays.
        self.h = [np.array(h[0]), np.array(h[1]), np.array(h[2])]

        # Node related properties.
        shape_nodes = (self.h[0].size+1, self.h[1].size+1, self.h[2].size+1)
        self.shape_nodes = shape_nodes
        self.nodes_x = np.r_[0., self.h[0].cumsum()] + self.origin[0]
        self.nodes_y = np.r_[0., self.h[1].cumsum()] + self.origin[1]
        self.nodes_z = np.r_[0., self.h[2].cumsum()] + self.origin[2]

        # Cell related properties.
        shape_cells = (self.h[0].size, self.h[1].size, self.h[2].size)
        self.shape_cells = shape_cells
        self.n_cells = np.prod(shape_cells)
        self.cell_centers_x = (self.nodes_x[1:] + self.nodes_x[:-1])/2
        self.cell_centers_y = (self.nodes_y[1:] + self.nodes_y[:-1])/2
        self.cell_centers_z = (self.nodes_z[1:] + self.nodes_z[:-1])/2

        # Edge related properties.
        self.shape_edges_x = (shape_cells[0], shape_nodes[1], shape_nodes[2])
        self.shape_edges_y = (shape_nodes[0], shape_cells[1], shape_nodes[2])
        self.shape_edges_z = (shape_nodes[0], shape_nodes[1], shape_cells[2])
        self.n_edges_x = np.prod(self.shape_edges_x)
        self.n_edges_y = np.prod(self.shape_edges_y)
        self.n_edges_z = np.prod(self.shape_edges_z)
        self.n_edges = self.n_edges_x + self.n_edges_y + self.n_edges_z

        # Face related properties.
        self.shape_faces_x = (shape_nodes[0], shape_cells[1], shape_cells[2])
        self.shape_faces_y = (shape_cells[0], shape_nodes[1], shape_cells[2])
        self.shape_faces_z = (shape_cells[0], shape_cells[1], shape_nodes[2])
        self.n_faces_x = np.prod(self.shape_faces_x)
        self.n_faces_y = np.prod(self.shape_faces_y)
        self.n_faces_z = np.prod(self.shape_faces_z)
        self.n_faces = self.n_faces_x + self.n_faces_y + self.n_faces_z

    def __repr__(self):
        """Simple representation."""
        return (f"TensorMesh: {self.shape_cells[0]} x {self.shape_cells[1]} x "
                f"{self.shape_cells[2]} ({self.n_cells:,})")

    @property
    def cell_volumes(self):
        """Construct cell volumes of the 3D model as 1D array."""
        if getattr(self, '_cell_volumes', None) is None:
            self._cell_volumes = (
                    self.h[0][None, None, :]*self.h[1][None, :, None] *
                    self.h[2][:, None, None]).ravel()
        return self._cell_volumes


@utils._known_class
class TensorMesh(discretize.TensorMesh if discretize else BaseMesh):
    """A slightly modified version of :class:`discretize.TensorMesh`.

    Adds a few custom attributes (``__eq__``, ``copy``, and ``{to;from}_dict``)
    to :class:`discretize.TensorMesh`.

    It falls back to a minimal :class:`emg3d.meshes.BaseMesh` if discretize is
    not installed. Nothing fancy is possible with the minimal TensorMesh,
    particularly *no* plotting.


    Parameters
    ----------
    h : [array_like, array_like, array_like]
        Cell widths in x, y, and z directions.

    origin : array_like
        Origin (x, y, z).

    """

    def __eq__(self, mesh):
        """Compare two meshes.

        The provided ``mesh`` can be either an emg3d or a discretize
        TensorMesh instance.

        """

        # Check if mesh is of the same instance.
        equal = mesh.__class__.__name__ == self.__class__.__name__

        # Check dimensions.
        if equal:
            equal *= len(mesh.shape_cells) == len(self.shape_cells)

        # Check shape.
        if equal:
            equal *= np.all(self.shape_cells == mesh.shape_cells)

        # Check distances and origin.
        if equal:
            equal *= np.allclose(self.h[0], mesh.h[0], atol=0)
            equal *= np.allclose(self.h[1], mesh.h[1], atol=0)
            equal *= np.allclose(self.h[2], mesh.h[2], atol=0)
            equal *= np.allclose(self.origin, mesh.origin, atol=0)

        return bool(equal)

    def copy(self):
        """Return a copy of the TensorMesh."""
        return self.from_dict(self.to_dict(True))

    def to_dict(self, copy=False):
        """Store the necessary information in a dict for serialization.

        Parameters
        ----------
        copy : bool, default: False
            If True, returns a deep copy of the dict.


        Returns
        -------
        out : dict
            Dictionary containing all information to re-create the TensorMesh.

        """
        out = {
            'hx': self.h[0],
            'hy': self.h[1],
            'hz': self.h[2],
            'origin': self.origin,
            '__class__': self.__class__.__name__
        }
        if copy:
            return deepcopy(out)
        else:
            return out

    @classmethod
    def from_dict(cls, inp):
        """Convert dictionary into :class:`emg3d.meshes.TensorMesh` instance.

        Parameters
        ----------
        inp : dict
            Dictionary as obtained from
            :func:`emg3d.meshes.TensorMesh.to_dict`. The dictionary needs the
            keys ``hx``, ``hy``, ``hz``, and ``origin``.

        Returns
        -------
        mesh : TensorMesh
            A :class:`emg3d.meshes.TensorMesh` instance.

        """
        inp = {k: v for k, v in inp.items() if k != '__class__'}
        return cls(h=[inp.pop('hx'), inp.pop('hy'), inp.pop('hz')], **inp)


def construct_mesh(frequency, properties, center, domain=None, vector=None,
                   seasurface=None, **kwargs):
    r"""Return a TensorMesh for given parameters.

    Designing an appropriate grid is the most time-consuming part of any 3D
    modelling:

    - Cell sizes should be small enough to represent changes in the model as
      well as to minimize interpolation errors in the fields.
    - The computational domain has to be big enough to avoid effects from the
      boundary condition.
    - The total number of cells should be small to speed up computation.

    These are in itself contradictory requirements, and they additionally
    depend all on the subsurface properties, the frequency under consideration,
    and the survey type.

    This function is a helper routine to construct an appropriate grid.
    However, there is no guarantee that it is the best or even a good grid. The
    constructed grid is frequency- and property-dependent. Some details are
    explained in other functions:

    - The minimum cell width :math:`\Delta_\text{min}` is a function of
      ``frequency``, ``properties[0]``, ``min_width_pps``, and
      ``min_width_limits``, see Equation :eq:`mincellwidth`.
    - The skin depth :math:`\delta` is a function of ``frequency`` and
      ``properties``, see Equation :eq:`skindepth`.
    - The wavelength :math:`\lambda` is a function of ``frequency`` and
      ``properties``, see Equation :eq:`wavelength`.

    The relation of the survey domain, computational domain, and buffer zone is
    shown in  :numref:`Figure %s <AutoGrid>` for a x-z-section; the y-direction
    behaves the same as the x-direction (the figures are only visible in the
    web version of the API docs on https://emg3d.emsig.xyz).

    .. figure:: ../_static/construct_mesh.png
       :align: center
       :alt: Sketch for automatic gridding.
       :name: AutoGrid

       Relation between survey domain (Ds, ``domain``), computational domain
       (Dc), and buffer zone. The survey domain should contain all sources (red
       stars) and receivers (blue triangles) as well as any important feature
       that should be represented in the data. The buffer zone is calculated as
       a function of wavelength with the provided property in the given
       direction.

    The buffer zone around the survey domain is by default one wavelength. This
    means that the signal has to travel two wavelengths to get from the end of
    the survey domain to the end of the computational domain and back. This
    approach is quite conservative and on the safe side. You can reduce the
    buffer thickness if you know what you are doing. There are three parameters
    which influence the thickness of the buffer for a given frequency:
    ``properties``, which is used to calculate the skin depth and the
    wavelength, ``lambda_factor`` (default is 1) which sets how many times the
    wavelength is the thickness of the buffer (relative factor), and
    ``max_buffer``, which is an absolute maximum for the buffer thickness. A
    graphical illustration is given in :numref:`Figure %s <Buffer>`.

    .. figure:: ../_static/construct_mesh2.png
       :align: center
       :alt: Sketch for the buffer zone.
       :name: Buffer

       The thickness of the buffer zone (B) for (I)
       ``lambda_from_center=False`` (default) and for (II)
       ``lambda_from_center=True``. The ``lambda_factor``
       (:math:`\lambda_{fact}`) is a simple scaling factor for the wavelength
       :math:`\lambda`. The ``max_buffer`` is an absolute limitation.


    Parameters
    ----------

    frequency : float
        Frequency (Hz) to calculate skin depth; both the minimum cell width and
        the extent of the buffer zone, and therefore of the computational
        domain, are a function of skin depth.

    properties : {float, array_like}
        Properties to calculate the skin depths, which in turn are used to
        calculate the minimum cell width (usually at source location) and the
        extent of the buffer around the survey domain. The properties can be
        either resistivities, conductivities, or the logarithm (natural or base
        10) thereof. By default it assumes resistivities, but it can be changed
        with the parameter ``mapping``.

        Five formats are recognized for properties: it can either be a float or
        a list of 2, 3, 4, or 7 floats. Depending on the format these
        properties are used to calculate the following parameters:

        - ``p``: min_width and buffer in all directions;

        - ``[p1, p2]``:

          - ``p1`` : min_width,
          - ``p2`` : buffer in all directions;

        - ``[p1, p2, p3]``:

          - ``p1`` : min_width,
          - ``p2`` : buffer in negative z-direction,
          - ``p3`` : buffer in all other directions;

        - ``[p1, p2, p3, p4]``:

          - ``p1`` : min_width,
          - ``p2`` : buffer in horizontal directions,
          - ``p3``; ``p4`` : buffer in negative; positive z-direction.

        - ``[p1, p2, p3, p4, p5, p6, p7]``:

          - ``p1`` : min_width,
          - ``p2``; ``p3`` : buffer in negative; positive x-direction,
          - ``p4``; ``p5`` : buffer in negative; positive y-direction,
          - ``p6``; ``p7`` : buffer in negative; positive z-direction.

    center : array_like
        Center coordinates (x, y, z). The mesh is centered around this point,
        which means that here is the smallest cell. Usually this is the source
        location.

    domain : {tuple, list, dict, None}, optional
        Contains the survey-domain limits. This domain should include all
        source and receiver positions as well as any important feature of the
        model. Format: ``([xmin, xmax], [ymin, ymax], [zmin, zmax])`` or
        ``{'x': [xmin, xmax], 'y': [ymin, ymax], 'z': [zmin, zmax]}``.

        It can be None, or individual lists can be None (e.g., ``(None, None,
        [zmin, zmax])``), in which case you have to provide either the
        corresponding ``distance`` or ``vector``, which is then assumed to span
        exactly the domain. If only one list is provided it is applied to all
        dimensions.

    distance : {tuple, list, dict, None}, optional
        An alternative to ``domain``: Instead of defining the domain in
        absolute values, they are defined here as distance from the center.
        Format:
        ``([xl, xr], [yl, yr], [zd, zu])`` or
        ``{'x': [xl, xr], 'y': [yl, yr], 'z': [zd, zu]}``.
        From this the domain is given as
        ``([cx-xl, cx+xr], [cy-yl, cy+yr], [cz-zd, cz+zu])``,
        where ``center=(cx, cy, cz)``.

    vector : {tuple, ndarray, dict, None}, optional
        Contains vectors of mesh-edges that should be used. If provided, the
        vector *must* at least include all of the survey domain. If ``domain``
        is not provided, it is defined as the minimum/maximum of the provided
        vector. Format: ``(xvector, yvector, zvector)`` or ``{'x': xvector,
        'y': yvector, 'z': zvector}``.

        It can be None, or individual ndarrays can be None (e.g., ``(xvector,
        yvector, None)``), in which case you have to provide a ``domain`` or
        ``distance``. If only one ndarray is provided it is applied to all
        dimensions.

    seasurface : float, default: None
        Air-sea interface. This has only to be set in the marine case, when the
        mesh in z-direction is sought for (and the interface is not contained
        in ``vector``). If set, it will ensure that at the sea surface is an
        actual boundary. It has to be bigger than the lower limit of the survey
        domain.

    stretching : {tuple, list, dict}, default: [1.0, 1.5]
        Maximum stretching factors in the form of ``[max Ds, max Dc]``: the
        first value is the maximum stretching for the survey domain (default is
        1.0), the second value is the maximum stretching for the buffer zone
        (default is 1.5). If a list is provided the same is used for all three
        dimension. Alternatively a tuple of three lists can be provided, ``(x,
        y, z)`` or ``{'x': x, 'y': y, 'z': z}``. Note that the first value has
        no influence on dimensions where a ``vector`` is provided.

    min_width_limits : {float, list, tuple, dict, None}, default: None
        Passed through to :func:`cell_width` as ``limits``. A tuple of three
        or a dict with ``x;y;z`` can be provided for direction dependent
        values. Note that this value has no influence on dimensions where a
        ``vector`` is provided.

    min_width_pps : {float, tuple, dict}, default: 3.0
        Passed through to :func:`cell_width` as ``pps``. A tuple of three or a
        dict with ``x;y;z`` can be provided for direction dependent values.
        Note that this value has no influence on dimensions where a ``vector``
        is provided.

    lambda_factor : float, default: 1.0
        The buffer is taken as one wavelength from the survey domain. This can
        be regarded as quite conservative (but safe). The parameter
        ``lambda_factor`` can be used to reduce (or increase) this factor.

    max_buffer : float, default: 100_000
        Maximum thickness of the buffer zone around survey domain. If
        ``lambda_from_center=True``, this is the maximum distance from the
        center to the end of the computational domain.

    lambda_from_center : bool, default: False
        Flag how to compute the extent of the computational mesh as a function
        of wavelength:

        - False (default): The distance from the edge of the survey domain to
          the edge of the computational domain is one wavelength.
        - True: The distance from the center to the edge of the computational
          domain and back to the end of the survey domain is two wavelengths.

    mapping : {str, Map}, default: 'Resistivity'
        Defines what type the input ``property_{x;y;z}``-values correspond to.
        By default, they represent resistivities (Ohm.m). The implemented
        mappings are:

        - ``'Resistivity'``; ρ (Ω m);
        - ``'Conductivity'``; σ (S/m);
        - ``'LgResistivity'``; log_10(ρ);
        - ``'LgConductivity'``; log_10(σ);
        - ``'LnResistivity'``; log_e(ρ);
        - ``'LnConductivity'``; log_e(σ).

    cell_numbers : array_like, optional
        List of possible numbers of cells. See :func:`good_mg_cell_nr`.
        Default is ``good_mg_cell_nr(1024, 5, 3)``, which corresponds to
        numbers 16, 24, 32, 40, 48, 64, 80, 96, 128, 160, 192, 256, 320, 384,
        512, 640, 768, 1024.

    verb : int, default: 0
        If 1 verbose, if 0 silent. The info is added either way to the returned
        mesh as ``mesh.construct_mesh_info``.


    Returns
    -------
    grid : TensorMesh
        Resulting mesh, a :class:`emg3d.meshes.TensorMesh` instance.

    """
    kwargs = deepcopy(kwargs)  # To not change a provided dict.
    verb = kwargs.get('verb', 0)
    distance = kwargs.pop('distance', None)

    # Initiate direction-specific dicts, add unambiguous args.
    kwargs['frequency'] = frequency
    kwargs['verb'] = -1            # Run x/y/z first, print/collect
    kwargs['raise_error'] = False  # info, then raise if necessary.
    xparams = {'center': center[0]}
    yparams = {'center': center[1]}
    zparams = {'center': center[2], 'seasurface': seasurface}

    # Add properties.
    if isinstance(properties, (int, float)):
        properties = np.array([properties])
    if len(properties) == 3:
        xparams['properties'] = [properties[0], properties[2], properties[2]]
        yparams['properties'] = [properties[0], properties[2], properties[2]]
        zparams['properties'] = [properties[0], properties[1], properties[2]]
    elif len(properties) == 4:
        xparams['properties'] = [properties[0], properties[1], properties[1]]
        yparams['properties'] = [properties[0], properties[1], properties[1]]
        zparams['properties'] = [properties[0], properties[2], properties[3]]
    elif len(properties) == 7:
        xparams['properties'] = [properties[0], properties[1], properties[2]]
        yparams['properties'] = [properties[0], properties[3], properties[4]]
        zparams['properties'] = [properties[0], properties[5], properties[6]]
    else:
        kwargs['properties'] = properties

    def _put_in_dicts(dicts, value, name):
        """Loop over dicts and put corresponding values."""
        for i, data in enumerate(dicts):
            if value[i] is not None:
                data[name] = value[i]

    # Add optionally direction specific args.
    for name, value in zip(['domain', 'vector', 'distance'],
                           [domain, vector, distance]):
        if value is None or isinstance(value, np.ndarray):
            kwargs[name] = value
        elif isinstance(value, dict):
            _put_in_dicts([xparams, yparams, zparams],
                          (value['x'], value['y'], value['z']), name)
        elif len(value) == 3:
            _put_in_dicts([xparams, yparams, zparams], value, name)
        else:
            kwargs[name] = value

    # Add optionally direction specific kwargs.
    for name in ['stretching', 'min_width_limits', 'min_width_pps']:
        value = kwargs.pop(name, None)
        if value is not None:
            if isinstance(value, (int, float)):
                kwargs[name] = np.array([value])
            elif isinstance(value, dict):
                _put_in_dicts([xparams, yparams, zparams],
                              (value['x'], value['y'], value['z']), name)
            elif len(value) == 3:
                _put_in_dicts([xparams, yparams, zparams], value, name)
            else:
                kwargs[name] = value

    # Get origins and widths in all directions.
    x0, hx, xinfo = origin_and_widths(**kwargs, **xparams)
    y0, hy, yinfo = origin_and_widths(**kwargs, **yparams)
    z0, hz, zinfo = origin_and_widths(**kwargs, **zparams)

    # Throw message if no solution was found.
    if any([out is None for out in [x0, y0, z0]]):
        raise RuntimeError("No suitable grid found; relax your criteria.")

    # Create mesh.
    mesh = TensorMesh(h=[hx, hy, hz], origin=np.array([x0, y0, z0]))

    # Collect info.
    info = (f"\n         == GRIDDING IN X ==\n{xinfo}\n"
            f"\n         == GRIDDING IN Y ==\n{yinfo}\n"
            f"\n         == GRIDDING IN Z ==\n{zinfo}\n")
    mesh.construct_mesh_info = info
    if verb > 0:
        print(info)

    return mesh


def origin_and_widths(frequency, properties, center, domain=None, vector=None,
                      seasurface=None, **kwargs):
    r"""Return origin and cell widths for given parameters.

    Calculate and return the starting point (origin) and cell widths given
    the input parameters. The output can be used as input (for one dimension)
    to create a :class:`emg3d.meshes.TensorMesh`. The function is used by
    :func:`construct_mesh` once in each direction.

    .. note::

        The parameters are described in :func:`construct_mesh`. Described here
        are only the differences. It is recommended to use directly
        :func:`construct_mesh`.


    Parameters
    ----------

    verb : int, default: 0
        If 1 verbose, if 0 silent, if -1 it returns the info as string instead
        of printing it.

    raise_error : bool, default: True
        If True, an error is raised if no suitable grid is found. Otherwise it
        just returns None's (used by ``construct_mesh``).

    Returns
    -------
    origin : float
        Origin of the mesh.

    widths : ndarray
        Cell widths of mesh.

    info : str, returned if verb<0
        Info.

    """
    # Get all kwargs.
    distance = kwargs.pop('distance', None)
    stretching = kwargs.pop('stretching', [1.0, 1.5])
    min_width_limits = kwargs.pop('min_width_limits', None)
    min_width_pps = kwargs.pop('min_width_pps', 3)
    lambda_factor = kwargs.pop('lambda_factor', 1.0)
    max_buffer = kwargs.pop('max_buffer', 100000)
    lambda_from_center = kwargs.pop('lambda_from_center', False)
    pmap = kwargs.pop('mapping', 'Resistivity')
    cell_numbers = kwargs.pop('cell_numbers', good_mg_cell_nr())
    raise_error = kwargs.pop('raise_error', True)
    verb = kwargs.pop('verb', 0)

    # Ensure no kwargs left.
    if kwargs:
        raise TypeError(f"Unexpected **kwargs: {list(kwargs.keys())}.")

    # Get property map from string.
    if isinstance(pmap, str):
        pmap = getattr(maps, 'Map'+pmap)()

    # Properties.
    cond = pmap.backward(np.array(properties, ndmin=1, dtype=float))
    cond_arr = np.array([
        cond[0], cond[min(cond.size-1, 1)], cond[min(cond.size-1, 2)]])

    # Get skin depth.
    skind = skin_depth(frequency, cond_arr)

    # Minimum cell width.
    dmin = cell_width(skind[0], min_width_pps, min_width_limits)

    # Survey domain: if not provided get from vector or distance.
    # Priority: domain > vector > distance.
    if domain is None and vector is None and distance is None:
        raise ValueError(
            "At least one of `domain`, `distance`, "
            "and `vector` must be provided."
        )
    elif domain is None:
        if vector is None:
            domain = np.array([center-abs(distance[0]),
                               center+abs(distance[1])])
        else:
            domain = np.array([vector.min(), vector.max()], dtype=float)
    else:
        domain = np.array(domain, dtype=np.float64)
        if vector is not None:
            if domain[0] < vector.min() or domain[1] > vector.max():
                raise ValueError(
                    "Provided vector MUST at least include "
                    "all of the survey domain."
                )

    # Seasurface related checks.
    if seasurface is not None:

        # Check that seasurface > center.
        if seasurface <= center:
            raise ValueError("The `seasurface` must be bigger than `center`.")

        # If center is close to seasurface, set it to seasurface.
        if abs(seasurface - center) < dmin:
            center = seasurface

    # Computation domain; big enough to avoid boundary effects.
    # To avoid boundary effects we want the signal to travel two wavelengths
    # from the source to the boundary and back to the receiver.
    # => 2*pi*sd ~ 6.3*sd = one wavelength => signal is ~ 0.2 %.
    # Two wavelengths we can safely assume it is zero.
    wlength = lambda_factor*wavelength(skind[1:])

    if lambda_from_center:
        # Center to edges of domain.
        in_domain = abs(domain - center)

        # Required buffer, additional to domain.
        d_buff = np.max([np.zeros(2), (2*wlength - in_domain)/2], axis=0)

        # Add buffer to domain.
        comp_domain = np.array([domain[0]-d_buff[0], domain[1]+d_buff[1]])

        # Restrict total domain to max_buffer.
        comp_domain[0] = max(comp_domain[0], center-max_buffer)
        comp_domain[1] = min(comp_domain[1], center+max_buffer)

    else:
        dbuffer = np.min([wlength, np.ones(2)*max_buffer], axis=0)
        comp_domain = np.array([domain[0]-dbuffer[0], domain[1]+dbuffer[1]])

    # Initiate flag if terminated.
    finished = False

    # Initiate alpha variables for survey and computation domains.
    sa, ca = 1.0, 1.0

    # Loop over possible cell numbers from small to big.
    for nx in np.unique(cell_numbers):

        # Loop over possible alphas for domain.
        for sa in np.linspace(1.0, stretching[0], 100):

            if vector is None:

                # Get current stretched grid cell sizes.
                thxl = dmin*sa**np.arange(nx)  # Left of origin.
                thxr = dmin*sa**np.arange(nx)  # Right of origin.

                # Adjust stretching for seasurface if required.
                if seasurface is not None and seasurface > center:
                    t_nx = np.r_[center, center+np.cumsum(thxr)]
                    ii = np.argmin(abs(t_nx-seasurface))
                    thxr[:ii] *= abs(seasurface-center)/np.sum(thxr[:ii])

                # Fill from center to left and right domain.
                nl = np.sum((center-np.cumsum(thxl)) > domain[0]) + 1
                nr = np.sum((center+np.cumsum(thxr)) < domain[1]) + 1

                # Create the current hx-array.
                hx = np.r_[thxl[:nl][::-1], thxr[:nr]]

                # Get actual domain:
                asurv_domain = [center - np.sum(thxl[:nl]),
                                center + np.sum(thxr[:nr])]

            else:
                # Store actual domain, current hx-array, and number of cells.
                asurv_domain = [vector[0], vector[-1]]
                hx = np.diff(vector)

            # Expand for seasurface if necessary.
            if seasurface is not None and seasurface > asurv_domain[-1]:
                thxr = hx[-1]*sa**np.arange(nx)
                sdepth = seasurface - asurv_domain[-1]

                # Get number of element, round down, and stretch.
                ii = np.argmax(np.cumsum(thxr) > sdepth)
                thxr = thxr[:ii]  # Restrict.
                thxr *= abs(seasurface-asurv_domain[-1])/np.sum(thxr)

                # Adjust actual domain, hx, and count.
                asurv_domain[1] += sum(thxr)
                hx = np.r_[hx, thxr]

            # Remaining number of cells
            nx_remain = nx - hx.size

            # Not good, try next.
            if nx_remain <= 0:
                continue

            # Store for verbosity
            hxo = hx

            # Loop over possible alphas for buffer.
            for ca in np.linspace(sa, stretching[1], 100):

                # Get current stretched grid cell sizes.
                thxl = hx[0]*ca**np.arange(1, nx_remain+1)   # Left of survey.
                thxr = hx[-1]*ca**np.arange(1, nx_remain+1)  # Right of survey.

                # Fill from survey to left and right domain.
                nl = np.sum((asurv_domain[0] - np.cumsum(thxl)) >
                            comp_domain[0]) + 1
                nr = np.sum((asurv_domain[1] + np.cumsum(thxr)) <
                            comp_domain[1]) + 1

                # Get remaining number of cells.
                nx_remain2 = nx_remain - nl - nr

                if nx_remain2 < 0:  # Not good, try next.
                    continue

                # Create hx-array.
                nl += int(np.floor(nx_remain2/2))  # If uneven, add one cell
                nr += int(np.ceil(nx_remain2/2))   # more on the right.
                hx = np.r_[thxl[:nl][::-1], hx, thxr[:nr]]

                # Compute origin.
                x0 = float(asurv_domain[0]-np.sum(thxl[:nl]))

                # Mark it as finished and break out of the loop.
                finished = True
                break

            if finished:
                break

        if finished:
            break

    # Raise Error or return Nones if not finish.
    if not finished:
        msg = "No suitable grid found; relax your criteria."
        # Throw message if no solution was found.
        if raise_error:
            raise RuntimeError(msg)
        else:
            x0, hx, info = None, None, msg

    else:  # Collect info about final grid.

        # Check max stretching.
        sa_adj = np.max([hxo[1:]/hxo[:-1], hxo[:-1]/hxo[1:]])
        sa_limit = min(1.5, stretching[0]+0.25)

        # Display precision.
        prec = int(np.ceil(max(0, -np.log10(min(hx))+1)))

        info = f"Skin depth     [m] : {skind[0]:.{prec}f}"
        if cond.size > 1:
            info += f" / {skind[1]:.{prec}f}"
        if cond.size > 2:
            info += f" / {skind[2]:.{prec}f}"
        info += "  [corr. to `properties`]\n"

        info += (
            f"Survey dom. DS [m] : "
            f"{domain[0]:.{prec}f} - {domain[1]:.{prec}f}\n"
            #
            f"Comp. dom. DC  [m] : {comp_domain[0]:.{prec}f} - "
            f"{comp_domain[1]:.{prec}f}\n"
            #
            f"Final extent   [m] : {x0:.{prec}f} - {x0+np.sum(hx):.{prec}f}\n"
            #
            f"Cell widths    [m] : {min(hxo):.{prec}f} / {max(hxo):.{prec}f} "
            f"/ {max(hx):.{prec}f}  [min(DS) / max(DS) / max(DC)]\n"
            #
            f"Number of cells    : {nx} ({hxo.size} / "
            f"{nx-hxo.size-nx_remain2} / {nx_remain2})  "
            f"[Total (DS/DC/remain)]\n"
            #
            f"Max stretching     : {sa:.3f} ({sa_adj:.3f}) / {ca:.3f}"
            "  [DS (seasurface) / DC]")

        if sa_adj > sa_limit:
            info += (f"\nNote: Stretching in DS >> {sa}.\nThe reason "
                     "is usually the interplay of center/domain/seasurface.")

    if verb > 0:
        print(info)

    # Return required info.
    if verb < 0:
        return x0, hx, info
    else:
        return x0, hx


def good_mg_cell_nr(max_nr=1024, max_prime=5, min_div=3):
    r"""Return 'good' cell numbers for the multigrid method.

    'Good' cell numbers are numbers which can be divided by two as many times
    as possible. At the end there should be a low prime number.

    The function adds all numbers

    .. math::

        p\ 2^n &\leq M \ , \text{ for}

        p &= {2, 3, ..., p_\text{max}} \ ;

        n &= {n_\text{min}, n_\text{min}+1, ..., \infty} \ ,

    where :math:`M, p_\text{max}, n_\text{min}` correspond to ``max_nr``,
    ``max_prime``, and ``min_div``, respectively.


    Parameters
    ----------
    max_nr : int, default: 1024
        Maximum number of cells.

    max_prime : int, default: 5
        Highest permitted prime number p for p*2^n. {2, 3, 5, 7} are good upper
        limits in order to avoid too big lowest grids in the multigrid method.

    min_div : int, default: 3
        Minimum times the number can be divided by two.


    Returns
    -------
    numbers : ndarray
        Array containing all possible cell numbers from lowest to highest.

    """
    # Primes till 20.
    primes = np.array([2, 3, 5, 7, 11, 13, 17, 19], dtype=np.int64)

    # Sanity check; 19 is already ridiculously high.
    if max_prime > primes[-1]:
        raise ValueError(
            f"Highest prime is {max_prime}, please use a value < 20."
        )

    # Restrict to max_prime.
    primes = primes[primes <= max_prime]

    # Get possible values.
    # Currently restricted to prime*2**30 (for prime=2 => 1,073,741,824 cells).
    numbers = primes[:, None]*2**np.arange(min_div, 30)

    # Get unique values.
    numbers = np.unique(numbers)

    # Restrict to max_nr and return.
    return numbers[numbers <= max_nr]


def skin_depth(frequency, conductivity, mu_r=1.0):
    r"""Return skin depth as a function of frequency and conductivity.

    The skin depth :math:`\delta` (m) is given by

    .. math::
        :label: skindepth

        \delta = \sqrt{\frac{2}{\omega\sigma\mu}}\ ,

    where :math:`\omega=2\pi f` is angular frequency of frequency :math:`f`
    (Hz), :math:`\sigma` is conductivity (S/m), and :math:`\mu=\mu_\rm{r}\mu_0`
    is magnetic permeability (H/m).


    Parameters
    ----------
    frequency : float
        Frequency (Hz).

    conductivity : float
        Conductivity (S/m).

    mu_r : float, default: 1.0
        Relative magnetic permeability (-).


    Returns
    -------
    skindepth : float
        Skin depth (m).

    """
    skindepth = 1/np.sqrt(np.pi*abs(frequency)*conductivity*mu_r*mu_0)

    # For Laplace-domain computations.
    if frequency < 0:
        skindepth /= np.sqrt(2*np.pi)

    return skindepth


def wavelength(skin_depth):
    r"""Return wavelength as a function of skin depth.

    The wavelength :math:`\lambda` (m) is given by

    .. math::
        :label: wavelength

        \lambda = 2\pi\delta\ ,

    where the skin depth :math:`\delta` is a function of frequency and
    conductivity and is given by :func:`skin_depth`, Equation :eq:`skindepth`.


    Parameters
    ----------
    skin_depth : {float, ndarray}
        Skin depth (m).


    Returns
    -------
    wavelength : {float, ndarray}
        Wavelength (m).

    """
    return 2*np.pi*skin_depth


def cell_width(skin_depth, pps=3, limits=None):
    r"""Return cell width as function of points per skin depth.

    The cell width :math:`\Delta` is defined by the desired points per skin
    depth,

    .. math::
        :label: mincellwidth

        \Delta = \Delta_\text{min} \le \frac{\delta}{\text{pps}} \le
                 \Delta_\text{max} \ ,

    to ensure that there are ``pps`` cells per skin depth (unless restricted by
    ``limits``).

    The skin depth :math:`\delta` is a function of frequency and conductivity
    and is given by :func:`skin_depth`, Equation :eq:`skindepth`.


    Parameters
    ----------
    skin_depth : float
        Skin depth (m).

    pps : float
        Points per skin depth.

    limits : {None, float, array_like}, default: None
        Limits on cell width:

        - ``None``: No limits.
        - ``float``: Simply returns the ``limits`` as cell width.
        - ``[min, max]``: Cell width is limited to this range.


    Returns
    -------
    cell_width : float
        Cell width (m).

    """
    # Calculate cell width.
    cell_width = skin_depth/pps

    # Respect user limits.
    if limits is not None:

        limits = np.array(limits, ndmin=1)

        if limits.size == 1:
            cell_width = limits  # Ignores skin depth and pps.

        else:
            cell_width = np.clip(cell_width, *limits)  # Restrict.

    return cell_width



"""
Everything to store electromagnetic material properties for the solver.
"""
# Copyright 2018-2021 The EMSiG community.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

from copy import deepcopy

import numpy as np
from scipy.constants import epsilon_0

from emg3d import maps, meshes, utils

__all__ = ['Model', 'VolumeModel']


# MODEL
@utils._known_class
class Model:
    r"""A model containing the electromagnetic properties of the Earth.

    A model provides the required model parameters to the solver. The x-, y-,
    and z-directed electrical properties are by default resistivities. However,
    they can be defined as resistivities, :math:`\rho\ (\Omega\,\mathrm{m})`,
    or conductivities, :math:`\sigma\ (\mathrm{S/m})`, either on a linear or on
    a logarithmic scale (log or ln), by choosing the appropriate ``mapping``.
    Relative magnetic permeability :math:`\mu_\mathrm{r}` is by default set to
    one and electric permittivity :math:`\varepsilon_\mathrm{r}` is by default
    set to zero, but they can also be provided (isotropically). Keep in mind
    that the multigrid method as implemented in emg3d works for the diffusive
    approximation. When the displacement part in Maxwell's equations becomes
    too dominant it will start to fail (high frequencies or very high electric
    permittivity).


    Parameters
    ----------
    grid : TensorMesh
        The grid; a :class:`emg3d.meshes.TensorMesh` instance.

    property_{x;y;z} : {None, array_like}, default: 1 (x), None (y, z)
        Electrical material property in x-, y-, and z-directions. The
        properties are stored as Fortran-ordered arrays with the shape given by
        ``grid.shape``. The provided value must be broadcastable to that shape.

        By default, property refers to electrical resistivity. However, this
        can be changed with an appropriate ``mapping`` (the internals of
        emg3d work, irrelevant of the map, with electrical conductivities).

        The properties have to be of finite value, bigger than zero (on linear
        scale). The four supported anisotropy cases are:

        - ``x;y=None;z=None``: isotropic (``y=z=x``);
        - ``x;y=None;z``: vertical transverse isotropy VTI (``y=x``);
        - ``x;y;z=None``: horizontal transverse isotropy HTI (``z=x``);
        - ``x;y;z``: triaxial  anisotropy.

        If a property is not initiated it cannot be set later on (e.g., if a
        VTI model is created, it is not possible to set ``property_y`` later
        on, instead, a new model has to be initiated).

    mu_r, epsilon_r : {None, array_like}, default: None
        Relative magnetic permeability (-) and relative electric permittivity
        (-), respectively, both isotropic. The properties are stored as
        Fortran-ordered arrays with the shape given by ``grid.shape``. The
        provided value must be broadcastable to that shape.

        The properties have to be of finite value, bigger than zero.

        The relative magnetic permeability is assumed to be 1 if not provided.

        The relative electric permittivity is assumed to be 0 if not provided,
        which ignores the displacement part completely (diffusive
        approximation)

    mapping : str, default: 'Resistivity'
        Defines what type the electrical input ``property_{x;y;z}``-values
        correspond to. The implemented types are:

        - ``'Resistivity'``; ρ (Ω m);
        - ``'Conductivity'``; σ (S/m);
        - ``'LgResistivity'``; log_10(ρ);
        - ``'LgConductivity'``; log_10(σ);
        - ``'LnResistivity'``; log_e(ρ);
        - ``'LnConductivity'``; log_e(σ).

    """

    def __init__(self, grid, property_x=1., property_y=None, property_z=None,
                 mu_r=None, epsilon_r=None, mapping='Resistivity'):
        """Initiate a new model."""

        # Store grid.
        self.grid = grid

        # Alias shape_cells and n_cells to shape and size.
        self.shape = self.grid.shape_cells
        self.size = self.grid.n_cells

        # Get and store map.
        self.map = getattr(maps, 'Map'+mapping)()

        # Initiate and store all parameters.
        self._property_x = self._init_parameter(property_x, 'property_x')
        self._property_y = self._init_parameter(property_y, 'property_y')
        self._property_z = self._init_parameter(property_z, 'property_z')
        self._mu_r = self._init_parameter(mu_r, 'mu_r',)
        self._epsilon_r = self._init_parameter(epsilon_r, 'epsilon_r')
        self._properties = ['property_x', 'property_y', 'property_z',
                            'mu_r', 'epsilon_r']

        # Store case.
        if self._property_y is None and self._property_z is None:
            self.case = 'isotropic'
        elif self._property_z is None:
            self.case = 'HTI'
        elif self._property_y is None:
            self.case = 'VTI'
        else:
            self.case = 'triaxial'

    def __repr__(self):
        """Simple representation."""
        return (f"{self.__class__.__name__}: {self.map.description}; "
                f"{self.case}{'' if self.mu_r is None else '; mu_r'}"
                f"{'' if self.epsilon_r is None else '; epsilon_r'}"
                f"; {self.shape[0]} x {self.shape[1]} x {self.shape[2]} "
                f"({self.size:,})")

    def __add__(self, model):
        """Add two models."""

        # Ensure model is a Model instance.
        if model.__class__.__name__ != 'Model':
            return NotImplemented

        # Check input.
        self._operator_test(model)

        # Apply operator.
        kwargs = self._apply_operator(model, np.add)

        # Return new Model instance.
        return Model(grid=self.grid, mapping=self.map.name, **kwargs)

    def __sub__(self, model):
        """Subtract two models."""

        # Ensure model is a Model instance.
        if model.__class__.__name__ != 'Model':
            return NotImplemented

        # Check input.
        self._operator_test(model)

        # Apply operator.
        kwargs = self._apply_operator(model, np.subtract)

        # Return new Model instance.
        return Model(grid=self.grid, mapping=self.map.name, **kwargs)

    def __eq__(self, model):
        """Compare two models."""

        # Check if model is a Model instance.
        equal = model.__class__.__name__ == 'Model'

        # Check input.
        if equal:
            try:
                self._operator_test(model)
            except ValueError:
                equal = False

        # Compare values if not None.
        if equal:
            for prop in self._properties:
                val = getattr(self, prop)
                if val is not None:
                    equal *= np.allclose(val, getattr(model, prop))

        return bool(equal)

    def copy(self):
        """Return a copy of the Model."""
        return self.from_dict(self.to_dict(True))

    def to_dict(self, copy=False):
        """Store the necessary information in a dict for serialization.

        Parameters
        ----------
        copy : bool, default: False
            If True, returns a deep copy of the dict.


        Returns
        -------
        out : dict
            Dictionary containing all information to re-create the Model.

        """
        out = {
            '__class__': self.__class__.__name__,  # v ensure emg3d-TensorMesh
            'grid': meshes.TensorMesh(self.grid.h, self.grid.origin).to_dict(),
            **{prop: getattr(self, prop) for prop in self._properties},
            'mapping': self.map.name,
        }
        if copy:
            return deepcopy(out)
        else:
            return out

    @classmethod
    def from_dict(cls, inp):
        """Convert dictionary into :class:`emg3d.models.Model` instance.

        Parameters
        ----------
        inp : dict
            Dictionary as obtained from :func:`emg3d.models.Model.to_dict`. The
            dictionary needs the keys ``property_x``, ``property_y``,
            ``property_z``, ``mu_r``, ``epsilon_r``, ``grid``, and ``mapping``;
            ``grid`` itself is also a dict which needs the keys ``hx``, ``hy``,
            ``hz``, and ``origin``.

        Returns
        -------
        model : Model
            A :class:`emg3d.models.Model` instance.

        """
        inp = {k: v for k, v in inp.items() if k != '__class__'}
        MeshClass = getattr(meshes, inp['grid']['__class__'])
        return cls(grid=MeshClass.from_dict(inp.pop('grid')), **inp)

    # ELECTRICAL PROPERTIES
    @property
    def property_x(self):
        r"""Electrical property in x-direction."""
        return self._property_x

    @property_x.setter
    def property_x(self, property_x):
        r"""Update electrical property in x-direction."""
        self._check_positive_finite(property_x, 'property_x')
        self._property_x[:] = np.asfortranarray(property_x, dtype=np.float64)

    @property
    def property_y(self):
        r"""Electrical property in y-direction."""
        return self._property_y

    @property_y.setter
    def property_y(self, property_y):
        r"""Update electrical property in y-direction."""
        self._check_positive_finite(property_y, 'property_y')
        self._property_y[:] = np.asfortranarray(property_y, dtype=np.float64)

    @property
    def property_z(self):
        r"""Electrical property in z-direction."""
        return self._property_z

    @property_z.setter
    def property_z(self, property_z):
        r"""Update electrical property in z-direction."""
        self._check_positive_finite(property_z, 'property_z')
        self._property_z[:] = np.asfortranarray(property_z, dtype=np.float64)

    @property
    def mu_r(self):
        r"""Relative magnetic permeability."""
        return self._mu_r

    @mu_r.setter
    def mu_r(self, mu_r):
        r"""Update relative magnetic permeability."""
        self._check_positive_finite(mu_r, 'mu_r')
        self._mu_r[:] = np.asfortranarray(mu_r, dtype=np.float64)

    @property
    def epsilon_r(self):
        r"""Relative electric permittivity."""
        return self._epsilon_r

    @epsilon_r.setter
    def epsilon_r(self, epsilon_r):
        r"""Update relative electric permittivity."""
        self._check_positive_finite(epsilon_r, 'epsilon_r')
        self._epsilon_r[:] = np.asfortranarray(epsilon_r, dtype=np.float64)

    # INTERPOLATION
    def interpolate_to_grid(self, grid, **interpolate_opts):
        """Interpolate the model to a new grid.


        Parameters
        ----------
        grid : TensorMesh
            Grid of the new model; a :class:`emg3d.meshes.TensorMesh` instance.

        interpolate_opts : dict
            Passed through to :func:`emg3d.maps.interpolate`. Defaults are
            ``method='volume'``, ``log=True``, and ``extrapolate=True``.


        Returns
        -------
        obj : Model
            A new :class:`emg3d.models.Model` instance on ``grid``.

        """

        # Get solver options, set to defaults if not provided.
        g2g_inp = {
            'method': 'volume',
            'extrapolate': True,
            'log': not self.map.name.startswith('L'),
            **({} if interpolate_opts is None else interpolate_opts),
            'grid': self.grid,
            'xi': grid,
        }

        # Interpolate property_{x;y;z}; mu_r; and epsilon_r; add to dict.
        model_inp = {}
        for prop in self._properties:
            var = getattr(self, prop)
            if var is None:
                model_inp[prop] = None
            else:
                model_inp[prop] = maps.interpolate(values=var, **g2g_inp)

        # Assemble new model.
        return Model(grid, mapping=self.map.name, **model_inp)

    # INTERNAL UTILITIES
    def _init_parameter(self, values, name):
        """Initiate parameter by casting and broadcasting."""

        # If None, exit.
        if values is None:
            return None

        # Cast it to an array of floats, in Fortran order.
        values = np.asfortranarray(values, dtype=np.float64)

        # If 1D array of self.size, reshape it.
        if values.size == self.size:
            values = values.reshape(self.shape, order='F')

        # If not of shape self.shape, broadcast it.
        elif values.shape != self.shape:
            values = np.ones(self.shape, order='F')*values

        # Check >0 and finite.
        self._check_positive_finite(values, name)

        return values

    def _check_positive_finite(self, values, name):
        """Check parameter values are positive (on linear scale) and finite."""

        # If it is None, it cannot be set.
        if hasattr(self, '_'+name) and getattr(self, '_'+name) is None:
            raise ValueError(
                f"Model was initiated without `{name}`; cannot set values."
            )

        # Get mapped values; checks are carried out on conductivities.
        if 'property_' in name:
            mapped = self.map.backward(np.asarray(values))
        else:
            mapped = values

        # Check they are positive.
        if not np.all(np.real(mapped) > 0.0):
            raise ValueError(f"`{name}` must be all bigger than zero.")

        # Check |val| < inf.
        if not np.all(np.isfinite(mapped)):
            raise ValueError(f"`{name}` must be all finite.")

    def _operator_test(self, model):
        """Check if ``self`` and ``model`` are consistent for operations."""

        # Ensure the two instances have the same grid.
        if self.grid != model.grid:
            raise ValueError("Models have different grids.")

        # Ensure the two instances have the same case.
        if self.case != model.case:
            raise ValueError("Models have different anisotropy.")

        # Ensure both or none has mu_r:
        if (self.mu_r is None) != (model.mu_r is None):
            raise ValueError("One model has mu_r, the other not.")

        # Ensure both or none has epsilon_r:
        if (self.epsilon_r is None) != (model.epsilon_r is None):
            raise ValueError("One model has epsilon_r, the other not.")

        # Ensure the two instances have the same mapping:
        if self.map.name != model.map.name:
            raise ValueError("Models have different mappings.")

    def _apply_operator(self, model, operator):
        """Apply the provided operator to self and model."""

        # Apply operator to property_{x;y;z}; mu_r; and epsilon_r; add to dict.
        kwargs = {}
        for prop in self._properties:
            val = getattr(self, prop)
            if val is None:
                kwargs[prop] = None
            else:
                kwargs[prop] = operator(val, getattr(model, prop))

        return kwargs


class VolumeModel:
    r"""Return simplified model with volume-averaged eta_{x;y;z}; zeta.

    Takes a model and a source field and returns the volume-averaged eta and
    zeta values. This is used internally by the solver.

    .. math::

        \eta_{\{x,y,z\}} = -V\mathrm{i}\omega\mu_0
              \left(\sigma_{\{x,y,z\}} + \mathrm{i}\omega\varepsilon\right)

    .. math::

        \zeta = V\mu_\mathrm{r}^{-1}


    Parameters
    ----------
    model : Model
        Model to transform to volume-averaged values.

    sfield : Field
       A VolumeModel is frequency-dependent. The frequency-information is taken
       from the provided source field.

    """

    def __init__(self, model, sfield):
        """Initiate a new model with volume-averaged properties."""

        # Store case and minimal TensorMesh.
        self.case = model.case
        self.grid = meshes.BaseMesh(model.grid.h, model.grid.origin)

        # Get volume for volume-averaged values.
        vol = self.grid.cell_volumes.reshape(model.shape, order='F')

        # Compute and store eta.
        for name in model._properties[:3]:

            prop = getattr(model, name)

            if prop is None:
                eta = None

            else:
                # Get conductivities.
                cond = model.map.backward(prop)

                # Diffusive approximation.
                if model.epsilon_r is None:
                    eta = -sfield.smu0*vol*cond

                # Complete version.
                else:
                    eta = -sfield.smu0*vol*(
                            cond + sfield.sval*epsilon_0*model.epsilon_r)

            setattr(self, '_eta_' + name[-1], eta)

        # Compute and store zeta.
        zeta = vol
        if model.mu_r is not None:
            zeta /= model.mu_r
        self._zeta = zeta

    @property
    def eta_x(self):
        r"""Volume-averaged eta in x-direction."""
        return self._eta_x

    @property
    def eta_y(self):
        r"""Volume-averaged eta in y-direction."""
        if self.case in ['HTI', 'triaxial']:
            return self._eta_y
        else:
            return self._eta_x

    @property
    def eta_z(self):
        r"""Volume-averaged eta in z-direction."""
        if self.case in ['VTI', 'triaxial']:
            return self._eta_z
        else:
            return self._eta_x

    @property
    def zeta(self):
        r"""Volume-averaged, isotropic zeta."""
        return self._zeta



"""
Everything that is related to fields: storing the electric field, obtaining the
magnetic field, generating the source field; obtaining the fields at receiver
locations.
"""
# Copyright 2018-2021 The EMSiG community.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import warnings
from copy import deepcopy

import numba as nb
import numpy as np
from scipy.constants import mu_0

from emg3d.core import _numba_setting
from emg3d import maps, meshes, models, utils, electrodes

__all__ = ['Field', 'get_source_field', 'get_receiver', 'get_magnetic_field']


@utils._known_class
class Field:
    r"""A Field contains the x-, y-, and z- directed electromagnetic fields.

    A Field is a simple container that has a 1D array ``Field.field``
    containing the x-, y-, and z-directed fields one after the other.
    The field can be any field, such as an electric field, a magnetic field,
    or a source field (which is an electric field).

    The particular fields can be accessed via the ``Field.f{x;y;z}``
    attributes, which are 3D arrays corresponding to the shape of the edges
    (electric fields) or the faces (magnetic fields) in this direction;
    sort-order is Fortran-like ('F').


    Parameters
    ----------

    grid : TensorMesh
        The grid; a :class:`emg3d.meshes.TensorMesh` instance.

    data : ndarray, default: None
        The actual data, a ``ndarray`` of size ``grid.n_edges`` (electric
        fields) or ``grid.n_faces`` (magnetic fields). If ``None``, it is
        initiated with zeros.

    frequency : float, default: None
        Field frequency (Hz), used to compute the Laplace parameter ``s``.
        Either positive or negative:

        - ``frequency > 0``: Frequency domain, hence
          :math:`s = \mathrm{i}\omega = 2\mathrm{i}\pi f` (complex);
        - ``frequency < 0``: Laplace domain, hence
          :math:`s = f` (real).

        This is primarily important for source fields. However, frequency
        information is also required to obtain the magnetic field from an
        electric field.

    dtype : dtype, default: complex
        Data type of the initiated field; only used if both ``frequency`` and
        ``data`` are None.

    electric : bool, default: True
        If electric, the properties live on the edges of the grid, if magnetic,
        they live on the faces.

    """

    def __init__(self, grid, data=None, frequency=None, dtype=None,
                 electric=True):
        """Initiate a new Field instance."""

        # Get dtype.
        if frequency is not None:  # Frequency is top priority.
            if frequency > 0:
                dtype = np.complex128
            elif frequency < 0:
                dtype = np.float64
            else:
                raise ValueError(
                    "`frequency` must be f>0 (frequency domain) or f<0 "
                    "(Laplace domain). Provided: {frequency} Hz."
                )
        elif data is not None:  # Data is second priority.
            dtype = data.dtype

        elif dtype is None:  # Default.
            dtype = np.complex128

        # Store grid, frequency, and electric.
        self.grid = grid
        self._frequency = frequency
        self.electric = electric

        # Store field.
        if data is None:
            field = np.zeros(self._get_prop('n'), dtype=dtype, order='F')
        else:
            field = np.asarray(data, dtype=dtype)
        self._field = utils.EMArray(field)

    def __repr__(self):
        """Simple representation."""
        return (f"{self.__class__.__name__}: "
                f"{['magnetic', 'electric'][self.electric]}; "
                f"{self.grid.shape_cells[0]} x {self.grid.shape_cells[1]} x "
                f"{self.grid.shape_cells[2]}; {self.field.size:,}")

    def __eq__(self, field):
        """Compare two fields."""
        equal = self.__class__.__name__ == field.__class__.__name__
        equal *= self.grid == field.grid
        equal *= self._frequency == field._frequency
        equal *= self.electric == field.electric
        if equal:
            equal *= np.allclose(self._field, field._field, atol=0, rtol=1e-10)
        return bool(equal)

    def copy(self):
        """Return a copy of the Field."""
        return self.from_dict(self.to_dict(copy=True))

    def to_dict(self, copy=False):
        """Store the necessary information of the Field in a dict.

        Parameters
        ----------
        copy : bool, default: False
            If True, returns a deep copy of the dict.


        Returns
        -------
        out : dict
            Dictionary containing all information to re-create the Field.

        """
        out = {
            '__class__': self.__class__.__name__,  # v ensure emg3d-TensorMesh
            'grid': meshes.TensorMesh(self.grid.h, self.grid.origin).to_dict(),
            'data': self._field,
            'frequency': self._frequency,
            'electric': self.electric,
        }
        if copy:
            return deepcopy(out)
        else:
            return out

    @classmethod
    def from_dict(cls, inp):
        """Convert dictionary into :class:`emg3d.fields.Field` instance.

        Parameters
        ----------
        inp : dict
            Dictionary as obtained from :func:`emg3d.fields.Field.to_dict`. The
            dictionary needs the keys ``field``, ``frequency``, ``grid``, and
            ``electric``; ``grid`` itself is also a dict which needs the keys
            ``hx``, ``hy``, ``hz``, and ``origin``.

        Returns
        -------
        field : Field
            A :class:`emg3d.fields.Field` instance.

        """
        inp = {k: v for k, v in inp.items() if k != '__class__'}
        MeshClass = getattr(meshes, inp['grid']['__class__'])
        return cls(grid=MeshClass.from_dict(inp.pop('grid')), **inp)

    @property
    def field(self):
        """Entire field as 1D array [fx, fy, fz]."""
        return self._field

    @field.setter
    def field(self, field):
        """Update field as 1D array [fx, fy, fz]."""
        self._field[:] = field

    @property
    def fx(self):
        """Field in x direction.

        Shape:

        - electric: (grid.cell_centers_x, grid.nodes_y, grid.nodes_z)
        - magnetic: (grid.nodes_x, grid.cell_centers_y, grid.cell_centers_z)

        """
        i1 = self._get_prop('n', 'x')
        shape = self._get_prop('shape', 'x')
        return self._field[:i1].reshape(shape, order='F')

    @fx.setter
    def fx(self, fx):
        """Update field in x-direction."""
        i1 = self._get_prop('n', 'x')
        self._field[:i1] = fx.ravel('F')

    @property
    def fy(self):
        """Field in y direction.

        Shape:

        - electric: (grid.nodes_x, grid.cell_centers_y, grid.nodes_z)
        - magnetic: (grid.cell_centers_x, grid.nodes_y, grid.cell_centers_z)

        """
        i0, i1 = self._get_prop('n', 'x'), self._get_prop('n', 'z')
        shape = self._get_prop('shape', 'y')
        return self._field[i0:-i1].reshape(shape, order='F')

    @fy.setter
    def fy(self, fy):
        """Update field in y-direction."""
        i0, i1 = self._get_prop('n', 'x'), self._get_prop('n', 'z')
        self._field[i0:-i1] = fy.ravel('F')

    @property
    def fz(self):
        """Field in z direction.

        Shape:

        - electric: (grid.nodes_x, grid.nodes_y, grid.cell_centers_z)
        - magnetic: (grid.cell_centers_x, grid.cell_centers_y, grid.nodes_z)

        """
        i0 = self._get_prop('n', 'z')
        shape = self._get_prop('shape', 'z')
        return self._field[-i0:].reshape(shape, order='F')

    @fz.setter
    def fz(self, fz):
        """Update electric field in z-direction."""
        i0 = self._get_prop('n', 'z')
        self._field[-i0:] = fz.ravel('F')

    @property
    def frequency(self):
        """Frequency (Hz)."""
        if self._frequency is None:
            return None
        else:
            return abs(self._frequency)

    @property
    def smu0(self):
        """s*mu_0; mu_0 = magn permeability of free space [H/m]."""
        if getattr(self, '_smu0', None) is None:
            if self.sval is not None:
                self._smu0 = self.sval*mu_0
            else:
                self._smu0 = None

        return self._smu0

    @property
    def sval(self):
        """Laplace parameter; s=iw in f-domain and s=f in Laplace-domain."""

        if getattr(self, '_sval', None) is None:
            if self._frequency is not None:
                if self._frequency < 0:  # Laplace domain; s.
                    self._sval = np.array(-self._frequency)
                else:  # Frequency domain; s = iw = 2i*pi*f.
                    self._sval = np.array(2j*np.pi*self._frequency)
            else:
                self._sval = None

        return self._sval

    def _get_prop(self, pre=None, post=None):
        """Returns `edges` or `faces` property depending on `electric`."""
        name = '' if pre is None else pre + '_'
        name += 'edges' if self.electric else 'faces'
        name += '' if post is None else '_' + post
        return getattr(self.grid, name)

    # INTERPOLATION
    def interpolate_to_grid(self, grid, **interpolate_opts):
        """Interpolate the field to a new grid.


        Parameters
        ----------
        grid : TensorMesh
            New grid; a :class:`emg3d.meshes.TensorMesh` instance.

        interpolate_opts : dict
            Passed through to :func:`emg3d.maps.interpolate`. Defaults are
            ``method='cubic'``, ``log=True``, and ``extrapolate=False``.


        Returns
        -------
        field : Field
            A new :class:`emg3d.fields.Field` instance on ``grid``.

        """

        # Get solver options, set to defaults if not provided.
        g2g_inp = {
            'method': 'cubic',
            'extrapolate': False,
            'log': True,
            **({} if interpolate_opts is None else interpolate_opts),
            'grid': self.grid,
            'xi': grid,
        }

        # Interpolate f{x;y;z}.
        field = np.r_[maps.interpolate(values=self.fx, **g2g_inp).ravel('F'),
                      maps.interpolate(values=self.fy, **g2g_inp).ravel('F'),
                      maps.interpolate(values=self.fz, **g2g_inp).ravel('F')]

        # Assemble and return new field.
        return Field(grid, field, frequency=self._frequency)

    def get_receiver(self, receiver):
        """Return the field (response) at receiver coordinates.

        Note that in order to avoid boundary effects from the PEC boundary the
        outermost cells are neglected. Field values for coordinates outside of
        the grid are set to NaN's. However, take into account that for good
        results all receivers should be far away from the boundary.

        Parameters
        ----------
        receiver : {Rx*, list, tuple}
            Receiver coordinates. The following formats are accepted:

            - ``Rx*`` instance, any receiver object from
              :mod:`emg3d.electrodes`.
            - ``list``: A list of ``Rx*`` instances.
            - ``tuple``: ``(x, y, z, azimuth, elevation)``; receiver
              coordinates and angles (m, °). All values can either be a scalar
              or having the same length as number of receivers.

            Note that the actual receiver type of the ``Rx*`` instances has no
            effect here, it just takes the coordinates from the receiver
            instances.


        Returns
        -------
        responses : EMArray
            Responses at receiver.

        """
        return get_receiver(self, receiver)


def get_source_field(grid, source, frequency, **kwargs):
    r"""Return source field for provided source and frequency.

    The source term is given in Equation 2 of [Muld06]_,

    .. math::

        -\mathrm{i} \omega \mu_0 \mathbf{J}_\mathrm{s} \, .

    The adjoint of the trilinear interpolation is used to distribute the points
    to the grid edges, which corresponds to the discretization of a Dirac
    ([PlDM07]_).


    Parameters
    ----------
    grid : TensorMesh
        Model grid; a :class:`emg3d.meshes.TensorMesh` instance.

    source : {Tx*, tuple, list, ndarray)
        Any source object from :mod:`emg3d.electrodes` (recommended usage).

        If it is a list, tuple, or ndarray it is in the case of a dipole put
        through to :class:`emg3d.electrodes.TxElectricDipole` or, if
        ``electric=False``, to :class:`emg3d.electrodes.TxMagneticDipole`. If
        it has more than two points it is put through to
        :class:`emg3d.electrodes.TxElectricWire`. Consult the documentation of
        the respective classes for the format, but it is recommended to provide
        directly the classes.

    frequency : float, default: None
        Source frequency (Hz), used to compute the Laplace parameter ``s``.
        Either positive or negative:

        - ``frequency > 0``: Frequency domain, hence
          :math:`s = \mathrm{i}\omega = 2\mathrm{i}\pi f` (complex);
        - ``frequency < 0``: Laplace domain, hence
          :math:`s = f` (real).

    strength : {float, complex}, default: 1.0
        Source strength (A), put through to
        :class:`emg3d.electrodes.TxElectricDipole` or, if ``electric=False``,
        to :class:`emg3d.electrodes.TxMagneticDipole`.

        | *Only used if the provided source is not a source instance.*

    length : float, default: 1.0
        Dipole length (m), put through to
        :class:`emg3d.electrodes.TxElectricDipole` or, if ``electric=False``,
        to :class:`emg3d.electrodes.TxMagneticDipole`.

        | *Only used if the provided source is not a source instance.*

    electric : bool, default: True
        If True, :class:`emg3d.electrodes.TxElectricDipole` is used to get the
        source instance, else :class:`emg3d.electrodes.TxMagneticDipole`.

        | *Only used if the provided source is not a source instance.*


    Returns
    -------
    sfield : Field
        Source field, a :class:`emg3d.fields.Field` instance.


    Examples
    --------

    .. ipython::

       In [1]: import emg3d
          ...: import numpy as np

       In [2]: # Create a simple grid, 8 cells of length 100 m in each
          ...: # direction, centered around the origin.
          ...: hx = np.ones(8)*100
          ...: grid = emg3d.TensorMesh([hx, hx, hx], origin=(-400, -400, -400))
          ...: grid  # For QC

       In [3]: # Create an electric dipole source from
          ...: # x1=y1=z1=0 to x2=100, y2=z2=0; strength=100 A.
          ...: source = emg3d.TxElectricDipole(
          ...:             [[0, 0, 0], [100, 0, 0]], strength=100)
          ...: source  # For QC

       In [4]: # Get the corresponding source field for f=0.5 Hz.
          ...: sfield = emg3d.get_source_field(grid, source, frequency=0.5)
          ...: sfield  # For QC

    """

    # Convert tuples, lists, and ndarrays to Source instances.
    if isinstance(source, (tuple, list, np.ndarray)):

        # Get optional kwargs and cast source.
        inp = {'strength': kwargs.get('strength', 1.0)}
        source = np.asarray(source)
        if source.size == 5:
            inp['length'] = kwargs.get('length', 1.0)

        # Call Tx*-class depending on provided source shape and `electric`.
        if source.size > 6:
            source = electrodes.TxElectricWire(source, **inp)
        elif kwargs.get('electric', True):
            source = electrodes.TxElectricDipole(source, **inp)
        else:
            source = electrodes.TxMagneticDipole(source, **inp)

    # Get total vector field by looping over segments.
    vfield = np.zeros(grid.n_edges, order='F')
    for p0, p1 in zip(source.points[:-1, :], source.points[1:, :]):

        # Add this segments' vector field to total vector field.
        vfield += _dipole_vector(grid, source=np.r_[[p0, p1]]).field

    # Initiate field with the total vector field.
    sfield = Field(grid, data=vfield, frequency=frequency)

    # Multiply by source strength
    sfield.field *= source.strength

    # Multiply by -i*w*mu_0 to obtain the full source term.
    sfield.field *= -sfield.smu0

    return sfield


def get_receiver(field, receiver):
    """Return the field (response) at receiver coordinates.

    Note that in order to avoid boundary effects from the PEC boundary the
    outermost cells are neglected. Field values for coordinates outside of the
    grid are set to NaN's. However, take into account that for good results all
    receivers should be far away from the boundary.


    Parameters
    ----------
    field : Field
        The electric or magnetic field; a :class:`emg3d.fields.Field` instance.

    receiver : {Rx*, list, tuple}
        Receiver coordinates. The following formats are accepted:

        - ``Rx*`` instance, any receiver object from :mod:`emg3d.electrodes`.
        - ``list``: A list of ``Rx*`` instances.
        - ``tuple``: ``(x, y, z, azimuth, elevation)``; receiver coordinates
          and angles (m, °). All values can either be a scalar or having the
          same length as number of receivers.

        Note that the actual receiver type of the ``Rx*`` instances has no
        effect here, it just takes the coordinates from the receiver instances.


    Returns
    -------
    responses : EMArray
        Responses at receiver.

    """

    # Rx* instance.
    if hasattr(receiver, 'coordinates'):
        coordinates = receiver.coordinates

    # List of Rx* instances.
    elif hasattr(tuple(receiver)[0], 'coordinates'):
        nrec = len(receiver)
        coordinates = np.zeros((nrec, 5))
        for i, r in enumerate(receiver):
            coordinates[i, :] = r.coordinates
        coordinates = tuple(coordinates.T)

    # Tuple of coordinates.
    else:
        coordinates = receiver

        # Check receiver dimension.
        if len(coordinates) != 5:
            raise ValueError(
                "`receiver` needs to be in the form "
                "(x, y, z, azimuth, elevation). "
                f"Length of provided `receiver`: {len(coordinates)}."
            )

    # Grid.
    grid = field.grid

    # Pre-allocate the response.
    _, xi, shape = maps._points_from_grids(
            grid, field.fx, coordinates[:3], 'cubic')
    resp = np.zeros(xi.shape[0], dtype=field.field.dtype)

    # Get weighting factors per direction.
    factors = electrodes.rotation(*coordinates[3:])

    # Add the required responses.
    opts = {'method': 'cubic', 'extrapolate': False, 'log': False, 'mode':
            'constant', 'cval': np.nan}
    for i, ff in enumerate((field.fx, field.fy, field.fz)):
        if np.any(abs(factors[i]) > 1e-10):
            resp += factors[i]*maps.interpolate(grid, ff, xi, **opts)

    # PEC: If receivers are in the outermost cell, set them to NaN.
    # Note: Receivers should be MUCH further away from the boundary.
    ind = ((xi[:, 0] < grid.nodes_x[1]) | (xi[:, 0] > grid.nodes_x[-2]) |
           (xi[:, 1] < grid.nodes_y[1]) | (xi[:, 1] > grid.nodes_y[-2]) |
           (xi[:, 2] < grid.nodes_z[1]) | (xi[:, 2] > grid.nodes_z[-2]))
    resp[ind] = np.nan

    # Return response.
    return utils.EMArray(resp.reshape(shape, order='F'))


def get_magnetic_field(model, efield):
    r"""Return magnetic field corresponding to provided electric field.

    Retrieve the magnetic field :math:`\mathbf{H}` from the electric field
    :math:`\mathbf{E}` using Farady's law, given by

    .. math::

        \nabla \times \mathbf{E} = \rm{i}\omega\mu\mathbf{H} .

    Note that the magnetic field is defined on the faces of the grid, unlike
    the electric field which is defined on the edges.


    Parameters
    ----------
    model : Model
        The model; a :class:`emg3d.models.Model` instance.

    efield : Field
        The electric field; a :class:`emg3d.fields.Field` instance.


    Returns
    -------
    hfield : Field
        The magnetic field; a :class:`emg3d.fields.Field` instance.

    """

    # Initiate magnetic field with zeros.
    hfield = Field(efield.grid, frequency=efield._frequency, electric=False)

    # Get volume-averaged mu_r divided by s*mu_0.
    vmodel = models.VolumeModel(model, efield)
    zeta = vmodel.zeta / efield.smu0

    # Compute magnetic field.
    _edge_curl_factor(
            hfield.fx[1:-1, :, :], hfield.fy[:, 1:-1, :],
            hfield.fz[:, :, 1:-1], efield.fx, efield.fy, efield.fz,
            efield.grid.h[0], efield.grid.h[1], efield.grid.h[2], zeta)

    return hfield


def _dipole_vector(grid, source, decimals=9):
    """Get unit dipole source vector using the adjoint interpolation method.


    Parameters
    ----------
    grid : TensorMesh
        The grid; a :class:`emg3d.meshes.TensorMesh` instance.

    source : ndarray
        Source coordinates of shape (2, 3): [[x1, y1, z1], [x2, y2, z2]] (m).
        Source field, a :class:`emg3d.fields.Field` instance; created if not
        provided.

    decimals : int, default: 9
        Grid nodes and source coordinates are rounded to given number of
        decimals. Default is nanometers.


    Returns
    -------
    vfield : Field
        Source field, a :class:`emg3d.fields.Field` instance.

    """
    vfield = Field(grid, dtype=float)

    # Round nodes and source coordinates (to avoid floating point issues etc).
    nodes_x = np.round(grid.nodes_x, decimals)
    nodes_y = np.round(grid.nodes_y, decimals)
    nodes_z = np.round(grid.nodes_z, decimals)
    source = np.round(np.asarray(source, dtype=float), decimals)

    # Ensure source is within nodes.
    outside = (
        min(source[:, 0]) < nodes_x[0] or max(source[:, 0]) > nodes_x[-1] or
        min(source[:, 1]) < nodes_y[0] or max(source[:, 1]) > nodes_y[-1] or
        min(source[:, 2]) < nodes_z[0] or max(source[:, 2]) > nodes_z[-1]
    )
    if outside:
        raise ValueError(f"Provided source outside grid: {source}.")

    # Dipole lengths in x-, y-, and z-directions, and overall.
    dxdydz = source[1, :] - source[0, :]
    length = np.linalg.norm(dxdydz)

    # Ensure finite length dipole is not a point dipole.
    if length < 1e-15:
        raise ValueError(f"Provided finite dipole has no length: {source}.")

    # Inverse source lengths.
    id_xyz = dxdydz.copy()
    id_xyz[id_xyz != 0] = 1/id_xyz[id_xyz != 0]

    # Cell fractions.
    a1 = (nodes_x - source[0, 0]) * id_xyz[0]
    a2 = (nodes_y - source[0, 1]) * id_xyz[1]
    a3 = (nodes_z - source[0, 2]) * id_xyz[2]

    # Get range of indices of cells in which source resides.
    def min_max_ind(vector, i):
        """Return [min, max]-index of cells in which source resides."""
        vmin = min(source[:, i])
        vmax = max(source[:, i])
        return [max(0, np.where(vmin < np.r_[vector, np.infty])[0][0]-1),
                max(0, np.where(vmax < np.r_[vector, np.infty])[0][0]-1)]

    rix = min_max_ind(nodes_x, 0)
    riy = min_max_ind(nodes_y, 1)
    riz = min_max_ind(nodes_z, 2)

    # Loop over these indices.
    for iz in range(riz[0], min(riz[1]+1, a3.size-1)):
        for iy in range(riy[0], min(riy[1]+1, a2.size-1)):
            for ix in range(rix[0], min(rix[1]+1, a1.size-1)):

                # Determine centre of gravity of line segment in cell.
                aa = np.vstack([[a1[ix], a1[ix+1]], [a2[iy], a2[iy+1]],
                                [a3[iz], a3[iz+1]]])
                aa = np.sort(aa[dxdydz != 0, :], 1)
                al = max(0, aa[:, 0].max())  # Left and right
                ar = min(1, aa[:, 1].min())  # elements.

                # Characteristics of this cell.
                xmin = source[0, :] + al*dxdydz
                xmax = source[0, :] + ar*dxdydz
                x_c = (xmin + xmax) / 2.0
                x_len = np.linalg.norm(xmax - xmin) / length

                # Contribution to edge (coordinate xyz)
                rx = (x_c[0] - nodes_x[ix]) / grid.h[0][ix]
                ex = 1 - rx
                ry = (x_c[1] - nodes_y[iy]) / grid.h[1][iy]
                ey = 1 - ry
                rz = (x_c[2] - nodes_z[iz]) / grid.h[2][iz]
                ez = 1 - rz

                # Add to field (only if segment inside cell).
                if min(rx, ry, rz) >= 0 and np.max(np.abs(ar-al)) > 0:

                    vfield.fx[ix, iy, iz] += ey*ez*x_len
                    vfield.fx[ix, iy+1, iz] += ry*ez*x_len
                    vfield.fx[ix, iy, iz+1] += ey*rz*x_len
                    vfield.fx[ix, iy+1, iz+1] += ry*rz*x_len

                    vfield.fy[ix, iy, iz] += ex*ez*x_len
                    vfield.fy[ix+1, iy, iz] += rx*ez*x_len
                    vfield.fy[ix, iy, iz+1] += ex*rz*x_len
                    vfield.fy[ix+1, iy, iz+1] += rx*rz*x_len

                    vfield.fz[ix, iy, iz] += ex*ey*x_len
                    vfield.fz[ix+1, iy, iz] += rx*ey*x_len
                    vfield.fz[ix, iy+1, iz] += ex*ry*x_len
                    vfield.fz[ix+1, iy+1, iz] += rx*ry*x_len

    # Ensure unity (should not be necessary).
    for field in [vfield.fx, vfield.fy, vfield.fz]:
        sum_s = abs(field.sum())
        if abs(sum_s-1) > 1e-6:
            # Print is always shown and simpler, warn for the CLI logs.
            msg = f"Normalizing Source: {sum_s:.10f}."
            print(f"* WARNING :: {msg}")
            warnings.warn(msg, UserWarning)
            field /= sum_s

    # Multiply by distance in each direction to obtain length.
    vfield.fx *= dxdydz[0]
    vfield.fy *= dxdydz[1]
    vfield.fz *= dxdydz[2]

    return vfield


@nb.njit(**_numba_setting)
def _edge_curl_factor(mx, my, mz, ex, ey, ez, hx, hy, hz, zeta):
    r"""Curl of values living on edges yielding values on faces.

    Used by :func:`emg3d.fields.get_magnetic_field` to compute the magnetic
    field from the electric field. The result is put into ``{mx;my;mz}``.


    Parameters
    ----------
    mx, my, mz : ndarray
        Pre-allocated zero magnetic field in x-, y-, and z-directions
        (:class:`emg3d.fields.Field`). The magnetic field grid has one cell
        less in each direction than the electric field grid.

    ex, ey, ez : ndarray
        Electric fields in x-, y-, and z-directions
        (:class:`emg3d.fields.Field`).

    hx, hy, hz : ndarray
        Cell widths in x-, y-, and z-directions
        (:class:`emg3d.meshes.TensorMesh`).

    zeta : ndarray
        Factor by which nabla x E will be divided. Shape of
        ``efield.grid.shape_cells``. In the case of Farady's law this is
        (smu)^-1, volume-averaged.

    """

    # Get dimensions
    nx = len(hx)
    ny = len(hy)
    nz = len(hz)

    # Loop over dimensions; x-fastest, then y, z
    for iz in range(nz):
        izm = max(0, iz-1)
        izp = iz+1
        for iy in range(ny):
            iym = max(0, iy-1)
            iyp = iy+1
            for ix in range(nx):
                ixm = max(0, ix-1)
                ixp = ix+1

                # Nabla x E.
                fx = ((ez[ix, iyp, iz] - ez[ix, iy, iz])/hy[iy] -
                      (ey[ix, iy, izp] - ey[ix, iy, iz])/hz[iz])
                fy = ((ex[ix, iy, izp] - ex[ix, iy, iz])/hz[iz] -
                      (ez[ixp, iy, iz] - ez[ix, iy, iz])/hx[ix])
                fz = ((ey[ixp, iy, iz] - ey[ix, iy, iz])/hx[ix] -
                      (ex[ix, iyp, iz] - ex[ix, iy, iz])/hy[iy])

                # Average zeta for dual-grid.
                dx = (hx[ixm] + hx[ix])/2
                dy = (hy[iym] + hy[iy])/2
                dz = (hz[izm] + hz[iz])/2
                zeta_x = (zeta[ixm, iy, iz] + zeta[ix, iy, iz])/2
                zeta_y = (zeta[ix, iym, iz] + zeta[ix, iy, iz])/2
                zeta_z = (zeta[ix, iy, izm] + zeta[ix, iy, iz])/2

                # Divide by zeta (averaged over the two cells) and store.
                if ix != 0:
                    mx[ixm, iy, iz] = fx * zeta_x / (dx * hy[iy] * hz[iz])
                if iy != 0:
                    my[ix, iym, iz] = fy * zeta_y / (hx[ix] * dy * hz[iz])
                if iz != 0:
                    mz[ix, iy, izm] = fz * zeta_z / (hx[ix] * hy[iy] * dz)
